local ffi = require 'ffi'
local C = ffi.C
ffi.cdef[[
typedef struct {
char* data;
int length, capacity;
} buffer;
void* malloc (int) ;
void* realloc (void*, int) ;
void free (void*) ;
]]

local proxy
proxy = {
__index = function (self, name)
local f = rawget(proxy,name)
if not f then
f = function (b, ...) return string[name](tostring(b), ...) end
rawset(proxy, name, f)
end
return f
end
}
local buffer = {
new = function (n)
local b = ffi.new('buffer')
b.length = 0
b.capacity = n
b.data = C.malloc(b.capacity)
return b
end,
append = function (b, str, len)
if type(str)~='string' then str=tostring(str) end
len = len or #str
if b.length+len+1 > b.capacity then
b.capacity = math.ceil(math.max(b.capacity * 3/2 +1, b.length+len+1))
b.data = C.realloc(b.data, b.capacity)
end
ffi.copy(b.data+b.length, str, len)
b.length = b.length + len
return b
end,
clear = function (b)
b.length=0
end,
tostring = function (b) 
return ffi.string(b.data, b.length)
end,
__len = function (b)
return buffer.length
end
}
buffer.__concat = buffer.append
buffer.__tostring = buffer.tostring
buffer.__index=buffer
setmetatable(buffer, proxy)
ffi.metatype('buffer', buffer)
return buffer
