local ljref = require 'ljrefs'
local ffi = require 'ffi'
local C, sdl = ffi.C, ffi.load('sdl')
ffi.cdef[[
typedef union{
  unsigned char type;

//   SDL_ActiveEvent active;
struct {
unsigned char type;
unsigned char gain;
unsigned char state;
} active;


// SDL_KeyboardEvent key;
struct{
unsigned char type;
unsigned char state;
struct {
unsigned char scancode;
int sym;
int mod;
unsigned short unicode;
} keysym;
} key;

//  SDL_MouseMotionEvent motion;
struct {
unsigned char type;
unsigned char state;
unsigned short x, y;
short xrel, yrel;
} motion;

//   SDL_MouseButtonEvent button;
struct {
unsigned char type;
unsigned char button;
unsigned char state;
unsigned short x, y;
} button;

//   SDL_JoyAxisEvent jaxis;
//   SDL_JoyBallEvent jball;
//   SDL_JoyHatEvent jhat;
//   SDL_JoyButtonEvent jbutton;
//   SDL_ResizeEvent resize;
//   SDL_ExposeEvent expose;
//   SDL_QuitEvent quit;
//   SDL_UserEvent user;
struct {
unsigned char type;
int code;
void *data1, *data2;
} user;
//   SDL_SywWMEvent syswm;
} SDL_Event;

bool wmWaitEvent (SDL_Event*, int) ;
bool wmPollEvent (SDL_Event*) ;
void wmPostEvent (SDL_Event*) ;
]]

local event = {
post = C.wmPostEvent,

none = 0,
keyDown = 2,
keyUp = 3,
mouseMoved = 4,
mouseDown = 5,
mouseUp = 6,
quit = 12,

active = {
lost = 0, gain = 1,
mouseFocus = 1, inputFocus = 2, active = 4
}
}

local function handlePrivateEvent (e)
if e.type==22 then
local t = ljref.getref(e.user.data1)
if e.user.code == 0 then ljref.unref(e.user.data1) end
t.f()
return false
end
return true
end

function event.new ()
return ffi.new('SDL_Event')
end

function event.poll (e)
e = e or oe event.new()
local re = C.wmPollEvent(e)
return re and handlePrivateEvent(e)
end

function event.wait (e, timeout)
e = e or event.new()
local re = C.wmWaitEvent(e, timeout or -1)
if re and not handlePrivateEvent(e) then return event.wait(e, timeout) end
return re 
end

setmetatable(event,event)
return event