local event = require 'event'
local window = require 'window'
local ffi = require 'ffi'
local C, sdl = ffi.C, ffi.load('sdl')
ffi.cdef[[
unsigned char* SDL_GetKeyState (void) ;
const char* SDL_GetKeyName (int key) ;
bool SDL_EnableUNICODE (int state) ;
bool SDL_EnableKeyRepeat (int delay, int interval) ;
]]

local key = {
backspace = 8, 
tab = 9, 
clear = 12, 
enter = 13,
pause = 19,
escape = 27,
space = 32,
exclaim = 33,
quote = 34,
sharp = 35, hash = 35,
dollar = 36,
enpersant = 38,
apos = 39, apostrophy = 39,
lparent = 40, leftParent = 40, rparent = 41, rightParent = 41,
asterisk = 42, 
plus = 43,
comma = 44,
minus = 45, dash = 45, 
period = 46, dot = 46,
slash = 47,
colon = 58,
semicolon = 59,
less = 60,
equal = 61,
greater = 62,
question = 63,
at = 64, arobase = 64,
lbracket = 91, leftBracket = 91,
backspace = 92, backSpace = 92,
rbracket = 93, rightBracket = 93,
vbar = 94, verticalBar = 94, caret = 94,
underscore = 95,
backquote = 96, backQuote = 96,
kpPeriod = 266, kpDot = 266,
kpDivide = 267,
kpMultiply = 268,
kpMinus = 269, kpDash = 269,
kpPlus = 270,
kpEnter = 271,
kpEqual = 272,
up = 273, upARrow = 273,
down = 274, downArrow = 274,
right = 275, rightArrow = 275,
left = 276, leftArrow = 276,
insert = 277,
home = 278,
['end'] = 279,
pageUp = 280,
pageDown = 281,
numLock = 300, numlock = 300,
capsLock = 301, capslock = 301,
scrollLock = 302,
rshift = 303, lshift = 304,
rctrl = 305, lctrl = 306,
ralt = 307, lalt = 308,
rmeta = 309, lmeta = 310,
rwin = 311, lwin = 312,
altgr = 313, compose = 314,
help = 315, print = 316, sysreq = 317,
['break'] = 318, 
menu = 319,
power = 320,
euro = 321,
undo = 322,
modifiers = {
lshift = 1, rshift = 2, shift = 3, 
lctrl = 0x40, rctrl = 0x80, ctrl = 0xC0,
lalt = 0x100, ralt = 0x200, alt = 0x300,
lmeta = 0x400, rmeta = 0x800, meta = 0xC00,
numLock = 0x1000,
capsLock = 0x2000,
altgr = 0x4000
}}
for i=0,9 do key['kp'..i] = 256+i end
for i=1,15 do key['f'..i] = 281+i end
for i=97,122 do key[string.char(i)] = i end

function key.name (k)
return ffi.string(sdl.SDL_GetKeyName(k))
end

function key.repeating (delay, interval)
delay = delay or 0
interval = interval or delay
return sdl.SDL_EnableKeyRepeat(delay, interval)
end

do
local state = false
function key.state (k)
if not state then state = sdl.SDL_GetKeyState() end
return state[k] ~= 0
end end

function key.pushed (k)
return key.state(k)
end

function key.released (k)
return not key.state(k)
end

function key.wait (e, timeout)
e = e or event.new()
local keyDown, wait = event.keyDown, event.wait
while wait(e, timeout) do
if e.type == keyDown then 
return e.key.keysym.sym, e.key.keysym.unicode 
elseif e.type == 21 then return
end
end
end

function key.__index (self, name)
if name == 'unicode' then return sdl.SDL_EnableUNICODE(-1)
end end

function key.__newindex (self, name, val) 
if name == 'unicode' then sdl.SDL_EnableUNICODE(val)
end end

setmetatable(key,key)
rawset(event,'key',key)
rawset(window,'key',key)
return key