local ffi = require 'ffi'
local C = ffi.C
ffi.cdef[[
double isfinite2 (double) ;
bool isnan2 (double) ;
bool isinf2 (double) ;
double lgamma2 (double) ;
double tgamma2 (double) ;
char* lltoa2 (long long num, char* buf, int base) ;
]]
math.gamma = C.tgamma2
math.lgamma = C.lgamma2
math.isnan = C.isnan2
math.isfinite = C.isfinite2
math.isinf = C.isinf2
math.e = math.exp(1)
math.nan = 0/0
math.inf = 1/0
math.ninf = -1/0

function math.isinteger (n)
return math.floor(n)==n
end

function math.round(n,p) 
p = 10^(p or 0)
  if n >= 0 then return math.floor(n*p+0.5)/p 
else return math.ceil(p*n-0.5)/p
end end

function math.fact (n)
return math.gamma(n+1)
end

do 
local log = math.log
function math.log (n, b)
return log(n)/log(b or math.e)
end
end

do
local nextval = nil
function math.randomGaussian ()
if nextval then
local x = nextval
nextval=nil
return x
end
local v1, v2, s
while not s or s>=1 or s==0 do
v1 = 2*math.random()-1 -- between -1 and 1
v2 = 2*math.random()-1 -- between -1 and 1
s = v1*v1 + v2*v2
end
local multiplier = math.sqrt( -2 * math.log(s)/s)
nextval = multiplier*v2
return v1 * multiplier
end end

function math.tostring (n, base)
local buf = ffi.new('char[?]', 65, 0)
return ffi.string(C.lltoa2(n, buf, base or 10))
end



return math