local event = require 'event'
local window = require 'window'
local ffi = require 'ffi'
local bit = require 'bit'
local C, sdl = ffi.C, ffi.load('sdl')
ffi.cdef[[
int SDL_GetMouseState (int* x, int* y) ;
int SDL_GetRelativeMouseState (int* x, int* y) ;
void SDL_WarpMouse (int x, int y) ;
]]

local mouse = {
move = sdl.SDL_WarpMouse,

left = 1,
middle = 2,
right = 3,
wheelUp = 4,
wheelDown = 5,
extra1 = 6,
extra2 = 7
}

function mouse.pushed (button)
return bit.band(sdl.SDL_GetMouseState(nil, nil), bit.lshift(button -1)) ~= 0
end

function mouse.released (button)
return not mouse.pushed(button)
end

function mouse.position ()
local x, y = ffi.new('int[1]'), ffi.new('int[1]')
sdl.SDL_GetMouseState(x,y)
return x[0], y[0]
end

function mouse.relativePosition ()
local x, y = ffi.new('int[1]'), ffi.new('int[1]')
sdl.SDL_GetRelativeMouseState(x,y)
return x[0], y[0]
end

function mouse.wait (e, timeout)
e = e or event.new()
local mouseDown, wait = event.mouseDown, event.wait
while wait(e, timeout) do
if e.type == mouseDown then 
return e.button.button, e.button.x, e.button.y
elseif e.type==21 then return
end
end
end

setmetatable(mouse,mouse)
rawset(event,'mouse',mouse)
rawset(window,'mouse',mouse)
return mouse