local ffi = require 'ffi'
local C, sdl = ffi.C, ffi.load('SDL_net')
ffi.cdef[[
typedef struct {
unsigned char _ip[4];
unsigned short _port;
} SDLAddr;
typedef struct { bool ready; } TCPSocket;
typedef struct { bool ready; } UDPSocket;
typedef struct SocketSet SocketSet;

unsigned short nbo2u16 (unsigned short nbo) ;
int SDLNet_Init (void) ;
void SDLNet_Quit (void) ;
const char* SDLNet_GetError (void) ;
int SDLNet_ResolveHost (SDLAddr* addr, const char* host, int port) ;
const char* SDLNet_ResolveIP (SDLAddr* addr) ;
TCPSocket* SDLNet_TCP_Open (SDLAddr* addr) ;
TCPSocket* SDLNet_TCP_Accept (TCPSocket* server) ;
void SDLNet_TCP_Close (TCPSocket* sock) ;
SDLAddr* SDLNet_TCP_GetPeerAddress (TCPSocket* sock) ;
int SDLNet_TCP_Send (TCPSocket* sock, const void* data, int len) ;
int SDLNet_TCP_Recv (TCPSocket* sock, void* buffer, int len) ;
int tcpRecvLine (TCPSocket* sock, char* buf, int max) ;
SocketSet* SDLNet_AllocSocketSet (int max) ;
void SDLNet_FreeSocketSet (SocketSet*) ;
int SDLNet_CheckSockets (SocketSet* set, unsigned long timeout) ;
int SDLNet_AddSocket (SocketSet* set, const void* sock) ;
int SDLNet_DelSocket (SocketSet* set, const void* sock) ;
]]

local tcp = {
close = sdl.SDLNet_TCP_Close,
__gc = sdl.SDLNet_TCP_Close
}

local udp = {}
local sdladdr = {}

local socketSet = {
new = sdl.SDLNet_AllocSocketSet,
insert = sdl.SDLNet_AddSocket,
remove = sdl.SDLNet_DelSocket,
free = sdl.SDLNet_FreeSocketSet,
__gc = sdl.SDLNet_FreeSocketSet
}

local net = {
init = sdl.SDLNet_Init,
free = sdl.SDLNet_Quit,
tcp = tcp, udp = udp, socketSet = socketSet
}

function sdladdr.__index (self, name)
if name == 'ip' then return string.format('%d.%d.%d.%d', self._ip[0], self._ip[1], self._ip[2], self._ip[3])
elseif name == 'host' then return ffi.string(sdl.SDLNet_ResolveIP(self))
elseif name=='port' then return C.nbo2u16(self._port)
end end

function udp.check (self)
return self.ready
end

function tcp.open (host, port)
local addr
if port then
addr = ffi.new('SDLAddr')
if sdl.SDLNet_ResolveHost(addr, host, port) ~= 0 then return nil end
else
addr = host
end
return sdl.SDLNet_TCP_Open(addr)
end

function tcp.readLine (self)
return self:read()
end

function tcp.send (self, buf, len) 
return sdl.SDLNet_TCP_Send(self, buf, len or #buf) 
end

function tcp.recv (self, num)
local buf = ffi.new('char[?]', num or 4096)
if not num then 
if C.tcpRecvLine(self, buf, 4096)>0 then return ffi.string(buf) end
else 
num = sdl.SDLNet_TCP_Recv(self, buf, num)
if num>0 then return ffi.string(buf, num)
end end
end

function tcp.lines (self)
return tcp.readLine, self
end

function tcp.check (self)
return self.ready
end

function tcp.__index (self, name)
if name=='peer' then return sdl.SDLNet_TCP_GetPeerAddress(self)
elseif name == 'ip' then return self.peer.ip
elseif name=='host' then return self.peer.host
elseif name=='port' then return self.peer.port
else return tcp[name]
end end

function socketSet.check (self, timeout) 
return sdl.SDLNet_CheckSockets(self, timeout or -1) 
end

tcp.write = tcp.send
tcp.read, tcp.receive = tcp.recv, tcp.rev
ffi.metatype('TCPSocket', tcp)
ffi.metatype('UDPSocket', udp)
ffi.metatype('SocketSet', socketSet)
ffi.metatype('SDLAddr', sdladdr)
return net