local ffi = require 'ffi'
local C, srapi = ffi.C, ffi.load('ScreenReaderAPI')
ffi.cdef[[
bool sayStringA (const char* str, bool interrupt) ;
bool stopSpeech (void) ;
bool setCurrentScreenReader (int id) ;
const char* getCurrentScreenReaderNameA (void) ;
bool setCurrentScreenReaderNameA (const char* name) ;
int getSupportedScreenReadersCount (void) ;
const char* getScreenReaderNameA (int num) ;
bool sapiIsEnabled (void) ;
bool sapiEnable (bool enable) ;
bool sapiSayStringA (const char* str, bool interrupt) ;
bool sapiStopSpeech (void) ;
bool sapiSetRate (int rate) ;
int sapiGetRate (void) ;
bool sapiSetVolume (int volume) ;
int sapiGetVolume (void) ;
bool sapiWait (int timeout) ;
bool sapiIsSpeaking (void) ;
int sapiGetNumVoices (void) ;
bool sapiSetVoice (int num) ;
int sapiGetVoice (void) ;
char* sapiGetVoiceNameA (int num) ;
bool sapiLoad (void) ;
bool sapiUnload (void) ;
void free (void*) ;
]]

local speech = {
stop = srapi.stopSpeech
}

local tts = {
init = srapi.sapiLoad,
free = srapi.sapiUnload,
stop = srapi.stopSpeech,
}

function speech.say (str, interrupt)
if interrupt==nil then interrupt=true end
return srapi.sayStringA(str, interrupt)
end

function speech.name (num)
return ffi.string(srapi.getScreenReaderNameA(num))
end

function speech.list ()
local t = {}
for i = 1, speech.count do
table.insert(t, speech.name(i))
end
return t
end

function speech.__index (self, name)
if name == 'current' then return ffi.string(srapi.getCurrentScreenReaderNameA())
elseif name == 'count' then
local count = srapi.getSupportedScreenReadersCount()
rawset(speech, 'count', count)
return count
end end

function speech.__newindex (self, name, val) 
if name == 'current' then
if type(val)=='string' then srapi.setCurrentScreenReaderNameA(val)
else srapi.setCurrentScreenReader(val)
end
end end

function tts.say (str, interrupt)
if interrupt==nil then interrupt=true end
return srapi.sapiSayStringA(str, interrupt)
end

function tts.wait (timeout)
return srapi.sapiWait(timeout or -1)
end

function tts.name (num)
local namePtr = srapi.sapiGetVoiceNameA(num -1)
if not namePtr then return nil end
local name = ffi.string(namePtr)
C.free(namePtr)
return name
end

function tts.list ()
local t = {}
for i = 1, tts.count do
table.insert(t, tts.name(i))
end
return t
end

function tts.__index (self, name)
if name == 'volume' then return srapi.sapiGetVolume()
elseif name == 'rate' then return srapi.sapiGetRate()
elseif name == 'enabled' then return srapi.sapiIsEnabled()
elseif name == 'speaking' then return sapiIsSpeaking()
elseif name == 'voice' then return srapi.sapiGetVoice() +1
elseif name == 'count' then
local count = srapi.sapiGetNumVoices()
rawset(tts, 'count', count)
return count
end end

function tts.__newindex (self, name, val)
if name == 'volume' then srapi.sapiSetVolume(val)
elseif name == 'rate' then srapi.sapiSetRate(val)
elseif name == 'enabled' then srapi.sapiEnable(val)
elseif name == 'voice' then srapi.sapiSetVoice(val -1)
end end

speech.tts = tts
setmetatable(tts,tts)
setmetatable(speech,speech)
return speech