local ffi = require 'ffi'
ffi.cdef[[
struct SDL_mutex* SDL_CreateMutex (void) ;
int SDL_mutexP (struct SDL_mutex*) ;
void SDL_mutexV (struct SDL_mutex*) ;
void SDL_DestroyMutex (struct SDL_mutex*) ;
struct SDL_cond* SDL_CreateCond (void) ;
void SDL_CondSignal (struct SDL_cond*) ;
void SDL_CondBroadcast (struct SDL_cond*) ;
int SDL_CondWait (struct SDL_cond*, struct SDL_mutex*) ;
int SDL_CondWaitTimeout (struct SDL_cond*, struct SDL_mutex*, unsigned int) ;
struct SDL_sem* SDL_CreateSemaphore (int) ;
int SDL_SemPost (struct SDL_sem*) ;
int SDL_SemWait (struct SDL_sem*) ;
int SDL_DestroySemaphore (struct SDL_sem*) ;
int SDL_SemValue (struct SDL_sem*) ;

struct ConditionVariableImpl {
struct SDL_cond* cond;
struct SDL_mutex* mutex;
};
]]
local sdl = ffi.load('sdl')
local mutex = { 
new = sdl.SDL_CreateMutex,
lock = sdl.SDL_mutexP, 
unlock = sdl.SDL_mutexV 
}
local semaphore = {
new = sdl.SDL_CreateSemaphore,
acquire = sdl.SDL_SemWait,
release = sdl.SDL_SemPost,
value = sdl.SDL_SemValue
}
local condition = {
signal = function (self) sdl.SDL_CondSignal(self.cond) end,
signalAll = function (self) sdl.SDL_CondBroadcast(self.cond) end,
wait = function (self, ms)
if not ms or ms<=0 then return 0==sdl.SDL_CondWait(self.cond, self.mutex)
else return 0==sdl.SDL_CondWaitTimeout(self.cond, self.mutex, ms)
end end,
new = function (m) 
local c = ffi.new('struct ConditionVariableImpl')
c.mutex = m
c.cond = sdl.SDL_CreateCond()
return c
end,
}
ffi.metatype('struct SDL_mutex', {__index=mutex, __gc=sdl.SDL_DestroyMutex})
ffi.metatype('struct SDL_sem', {__index=semaphore, __gc = sdl.SDL_DestroySemaphore})
ffi.metatype('struct ConditionVariableImpl', {__index=condition, __gc=function (self) sdl.DestroyCond(self.cond) end})

function mxpcall (f, r, m, ...)
m:lock()
local t = {xpcall(f, r, ...)}
m:unlock()
return unpack(t)
end

function mpcall (f, m, ...)
m:lock()
local t = {pcall(f, ...)}
m:unlock()
return unpack(t)
end

return {
condition=condition,
semaphore=semaphore,
mutex=mutex,
lock=mutex
}
