local ffi = require 'ffi'
ffi.cdef[[
typedef union {
float v[3];
struct { float x, y, z; };
} vector;
]]
local vector = { }
vector.__index=vector


function vector.new (x, y, z)
local v = ffi.new('vector')
if y then
v.x = x or 0
v.y = y or 0
v.z = z or 0
elseif type(x)=='table' or type(x)=='cdata' or type(x)=='userdata' then
v.x = x.x or 0
v.y = x.y or 0
v.z = x.z or 0
end
return v
end

function vector.points (self)
return self.x, self.y, self.z
end

function vector.__add (v1, v2)
return vector.new(v1.x+v2.x, v1.y+v2.y, v1.z+v2.z)
end

function vector.__sub (v1, v2)
return vector.new(v1.x-v2.x, v1.y-v2.y, v1.z-v2.z)
end

function vector.__mul (v1, v2)
if type(v1)=='number' then return vector.new(v1*v2.x, v1*v2.y, v1*v2.z)
elseif type(v2)=='number' then return vector.new(v1.x*v2, v1.y*v2, v1.z*v2)
else return vector.new(v1.y*v2.z-v1.z*v2.y, v1.z*v2.x-v2.z*v1.x, v1.x*v2.y-v1.y*v2.x)
end end

function vector.__len (v)
return math.sqrt(v.x*v.x + v.y*v.y + v.z*v.z)
end

function vector.__unm (v)
return vector.new(-v.x, -v.y, -v.z)
end

function vector.__tostring (self)
return string.format("(%g;%g;%g)", self.x, self.y, self.z)
end

ffi.metatype('vector', vector)
return vector
