local event = require 'event'
local ljref = require 'ljrefs'
local bit = require 'bit'
local ffi = require 'ffi'
local C, sdl = ffi.C, ffi.load('sdl')
ffi.cdef[[
void wmPostQuitEvent (void) ;
void* wmSetTimer (int timeout, void* param, bool once) ;
void alert (const char* title, const char* msg) ;
bool confirm (const char* title, const char* msg) ;
char* prompt (const char* title, const char* msg, const char* text) ;
void free (void*) ;
int SDL_Init (int) ;
int SDL_WasInit (int) ;
void SDL_Delay (int ms) ;
int SDL_GetTicks (void) ;
void* SDL_SetVideoMode (int width, int height, int bpp, int flags) ;
void SDL_WM_SetCaption (const char* title, const char* icon) ;
bool SDL_WM_IconifyWindow (void) ;
int SDL_WM_GrabInput (int) ;
int SDL_GetAppState (void) ;
void SDL_RemoveTimer (void*) ;
void SDL_Quit (void) ;
]]

local window = {
alert = C.alert,
confirm = C.confirm,
init = function () return sdl.SDL_Init(0x221) == 0 end,
free = sdl.SDL_Quit,
grab = sdl.SDL_WM_GrabInput,
iconify = sdl.SDL_WM_IconifyWindow,
sleep = sdl.SDL_Delay,
wait = sdl.SDL_Delay,
time = sdl.SDL_GetTicks,
close = C.wmPostQuitEvent,
event = event
}

local function setTimeout (timeout, func, once)
local t = { f = func }
t.timerId = C.wmSetTimer(timeout, ljref.ref(t), once)
return t
end

function window.open (title, width, height)
if sdl.SDL_WasInit(0x20) ~= 0 then window.init() end
title = title or 'Untitled window'
width = width or 640
height = height or 480
sdl.SDL_WM_SetCaption(title, nil)
local surface = sdl.SDL_SetVideoMode(width, height, 0, 0)
if not surface then error('Unable to open window', 2) end
sdl.SDL_WM_SetCaption(title, nil)
rawset(window,'__title',title)
rawset(window, 'surface', surface)
end

function window.setTimeout (timeout, func)
return setTimeout(timeout, func, true) 
end

function window.setInterval (timeout, func)
return setTimeout(timeout, func, false) 
end

function window.clearTimeout (t)
C.SDL_RemoveTimer(t.timerId)
ljref.unref(t)
end
window.clearInterval = window.clearTimeout

function window.prompt (title, msg, text)
text = text or ''
local result = C.prompt(title, msg, text)
if result==nil then return nil end
local str = ffi.string(result)
C.free(result)
return str
end

function window.__index (self, name)
if name == 'title' then return rawget(window,'__title')
elseif name == 'state' then return sdl.SDL_GetAppState()
elseif name == 'active' then return bit.band(self.state, bit.bor(event.active.active, event.active.inputFocus)) == bit.bor(event.active.active, event.active.inputFocus)
elseif name == 'iconified' then return bit.band(self.state, event.active.active) == 0
elseif name == 'hasInput' then return bit.band(self.state, event.active.inputFocus) ~= 0
elseif name == 'hasMouse' then return bit.band(self.state, event.active.mouseFocus) ~= 0
end end

function window.__newindex (self, name, val)
if name == 'title' then 
sdl.SDL_WM_SetCaption(val, nil)
rawset(window,'__title',val)
end end

setmetatable(window,window)
return window