#include<stdlib.h>
#include<stdio.h>
#include<string.h>

#define LZQ_MAGIC (*(unsigned long long*)"LZQ1.0.0")

#define error(x) { printf("Failed at %s:%d\r\n", __FUNCTION__, __LINE__); return x; }

typedef struct {
unsigned long offset;
unsigned long long magic;
} LZQFileHeader;

static int ceil16 (int n) {
int m = n/16*16;
if (n==m) return m;
else return m+16;
}

int lzqOpenRead (FILE* fp, int* offset) {
if (!fp) error(-1)
LZQFileHeader h;
if (fseek(fp, -sizeof(h), SEEK_END)) error(-4)
if (1!=fread(&h, sizeof(h), 1, fp)) error(-2)
if (h.magic!=LZQ_MAGIC) error(-3)
if (offset) *offset = h.offset;
return 0;
}

void lzqCloseWrite (FILE* fp, int offset) {
LZQFileHeader h;
h.offset = offset;
h.magic = LZQ_MAGIC;
return 1!=fwrite(&h, sizeof(h), 1, fp);
}

char* lzqReadData (FILE* fp, int offset, int* length, const char* key, int keylen) {
if (fseek(fp, offset, SEEK_SET)) error(NULL)
unsigned long compressedSize;
if (1!=fread(&compressedSize, sizeof(unsigned long), 1, fp)) error(NULL)
char* cbuf = malloc(compressedSize);
if (!cbuf) error(NULL)
char* dbuf = NULL;
int uncompressedSize;
if (1!=fread(cbuf, compressedSize, 1, fp)) { free(cbuf); error(NULL) }
if (key && keylen>0 && decrypt(cbuf, cbuf, compressedSize, key, keylen)) { free(cbuf); error(NULL) }
uncompressedSize = uncompress(cbuf, compressedSize, NULL, 0);
dbuf = malloc(uncompressedSize);
if (!dbuf) { free(cbuf); error(NULL) }
if (uncompress(cbuf, compressedSize, dbuf, uncompressedSize)) { free(cbuf); free(dbuf); error(NULL) }
free(cbuf);
if (length) *length = uncompressedSize;
return dbuf;
}

int lzqWriteData (FILE* fp, int* offset, const char* data, unsigned long length, const char* key, int keylen) {
int compressedLength = length+40;
char* cbuf = malloc(compressedLength);
if (!cbuf) error(-1)
if (compress(data, length, cbuf, &compressedLength)) { free(cbuf); error(-2) }
if (key && keylen>0 && encrypt(cbuf, cbuf, compressedLength, key, keylen)) { free(cbuf); error(-5) }
int of = ftell(fp);
if (offset) *offset=of;
if (!fwrite(&compressedLength, sizeof(int), 1, fp) ||
!fwrite(cbuf, compressedLength, 1, fp) ) { free(cbuf); error(-3) }
free(cbuf);
return compressedLength;
}

int lzqReadEntry (char* buf, int bof, int* offset, char* name) {
if (offset) *offset = *(int*)(buf+bof);
bof += sizeof(int);
unsigned char namelength = *(unsigned char*)(buf+bof);
bof++;
strncpy(name, buf+bof, namelength);
name[namelength]=0;
return bof+namelength;
}

char* lzqWriteEntry (char* buf, int* bufpos, int* buflen, int offset, const char* name, int namelength) {
if (namelength<=0) namelength=strlen(name);
int max = *bufpos+namelength+6;
if (!buf || *buflen<=max) {
int max2 = *buflen *3/2 +2;
if (max2>max) max=max2;
buf = realloc(buf, max);
*buflen = max;
}
*(int*)(buf+(*bufpos)) = offset;
*bufpos += sizeof(int);
*(unsigned char*)(buf+(*bufpos)) = namelength;
(*bufpos)++;
memcpy(buf+(*bufpos), name, namelength);
(*bufpos)+= namelength ;
return buf;
}

/*int main (int argc, char** argv) {
int i;
const char input1[] = "Hello, world !";
const char input2[] = "How are you ? I'm fine !";
char name1[256]={0}, name2[256]={0};
char key[32];
char* buf = NULL;
char *output1=NULL, *output2=NULL;
int offset1, offset2, length1, length2, re, bof, blen;

for (i=0; i<32; i++) key[i]=i*i;

FILE* fp = fopen("test.lzq", "wb");
re = lzqWriteData(fp, &offset1, input1, sizeof(input1), key);
printf("lzqWriteData: re=%d, offset=%d\r\n", re, offset1);
re = lzqWriteData(fp, &offset2, input2, sizeof(input2), key);
printf("lzqWriteData: re=%d, offset=%d\r\n", re, offset2);
bof = 0;
blen=0;
buf = lzqWriteEntry(buf, &bof, &blen, offset1, "First entry", 0);
buf = lzqWriteEntry(buf, &bof, &blen, offset2, "Second entry", 0);
re = lzqWriteData(fp, &blen, buf, bof, key);
free(buf); buf=NULL;
printf("lzqWriteData: re=%d, offset=%d\r\n", re, blen);
lzqCloseWrite(fp, blen);
fclose(fp);

fp = fopen("test.lzq", "rb");
re = lzqOpenRead(fp, &bof);
printf("lzqOpenRead: re=%d, offset=%d\r\n", re, bof);
buf = lzqReadData(fp, bof, &blen, key);
printf("lzqReadData: buf=%p, length=%d\r\n", buf, blen);

bof=0;
bof = lzqReadEntry(buf, bof, &offset1, name1);
printf("#1: name=%s, offset=%d, bof=%d\r\n", name1, offset1, bof);
bof = lzqReadEntry(buf, bof, &offset2, name2);
printf("#2: name=%s, offset=%d, bof=%d\r\n", name2, offset2, bof);

output2 = lzqReadData(fp, offset2, &length2, key);
printf("lzqReadData: name=%s, offset=%d, length=%d, data=%p ", name2, offset2, length2, output2);
printf("%s\r\n", output2);
output1 = lzqReadData(fp, offset1, &length1, key);
printf("lzReadData: name=%s, offset=%d, length=%d, data=%p ", name1, offset1, length1, output1);
printf("%s\r\n", output1);
fclose(fp);
//free(output1);
//free(output2);
//free(buf);
return 0;
}*/
