#include "audio.h"
#include "sndfile.h"
#include<stdio.h>

static int sfRead (SNDFILE* sf, int format, float* buf, int len) {
switch(format){
case FORMAT_S16: return sf_read_short(sf, buf, len);
case FORMAT_FLOAT : return sf_read_float(sf, buf, len);
}}

static int sfSeek (SNDFILE* sf, int pos, int ance) {
return sf_seek(sf, pos, ance);
}

static void* sfClose (SNDFILE* sf, void* udata2, audioIO* io) {
sf_close(sf);
return NULL;
}

static void* sfFill (SNDFILE* sf, audioIO* io, audioInfo* info, SF_INFO* sfinfo) {
if (!sf) return NULL;
info->samplerate = sfinfo->samplerate;
info->channels = sfinfo->channels;
info->frames = sfinfo->frames;
info->format = sfinfo->format;
io->read = sfRead;
io->seek = sfSeek;
io->close = sfClose;
return sf;
}

void* audioIOOpen (void* handle, audioInfo* info, audioIO* io) {
int format = info->format;
SF_INFO sfinfo = { .samplerate=info->samplerate, .channels=info->channels, .frames=info->frames, .format=format };
SF_VIRTUAL_IO vio;
SNDFILE* sf = sf_open_virtual(&vio, SFM_READ, &sfinfo, handle);
return sfFill(sf, io, info, &sfinfo);
}

void* audioIOOpenFile (const char* file, audioInfo* info, audioIO* io) {
int format = info->format;
SF_INFO sfinfo = { .samplerate=info->samplerate, .channels=info->channels, .frames=info->frames, .format=format };
SNDFILE* sf = sf_open(file, SFM_READ, &sfinfo);
return sfFill(sf, io, info, &sfinfo);
}


typedef struct {
void* handle;
int pos;
void*(*close)(void*,void*,audioIO*);
int(*read)(void*,int,void*,int);
} DSP;

void* dspClose (DSP* dsp, void* u2, audioIO* io) {
if (!io || !u2) dsp->close(dsp->handle, NULL, NULL);
else if (dsp==u2) {
io->read = dsp->read;
io->close = dsp->close;
return dsp->handle;
}
else dsp->handle = dsp->close(dsp, u2, io);
if (dsp==u2 || !u2) free(dsp);
}
int dspRead (DSP* dsp, int format, void* buf, int num) {
if (format==FORMAT_FLOAT) {
float* fbuf = (float*)buf;
num = dsp->read(dsp->handle, FORMAT_FLOAT, buf, num);
for (int i=0; i<num; i++) {
//buf[i] *= sin(3.14159265 * 2.0 * 2.0 * dsp->pos++ /44100.0);
int signum = fbuf[i]<0? -1 : 1;
fbuf[i] = pow(signum*fbuf[i],0.5) * signum;
}
return num;
}
else {
short* sbuf = (short*)buf;
float fbuf[num];
num = dspRead(dsp, FORMAT_FLOAT, fbuf, num);
for (int i=0; i<num; i++) sbuf[i] = fbuf[i] * 32767;
return num;
}
}
void* dspAttach (void* handle, audioIO* io) {
DSP* dsp = malloc(sizeof(DSP));
dsp->pos = 0;
dsp->handle = handle;
dsp->close = io->close;
dsp->read = io->read;
io->read = dspRead;
io->close = dspClose;
return dsp;
}


