#include "global.h"
#include<string.h>
#include<lua/lua.h>
#include<lua/lualib.h>
#include<lua/lauxlib.h>

typedef struct {
char* buf;
int len, cap;
} buffer;

buffer* buffer_init (buffer* b, int size) {
b->len = 0;
b->cap = size;
b->buf = malloc(size);
}

buffer* export buffer_new (int size) {
buffer* b = malloc(sizeof(buffer));
return buffer_init(b,size);
}

buffer* export buffer_clear (buffer* b) {
b->len = 0;
return b;
}

void export buffer_free (buffer* b, int freeall) {
if (b->buf) free(b->buf);
if (freeall) free(b);
}

buffer* export buffer_append (buffer* b, const char* str, int len) {
if (len<=0) len = strlen(str);
if (b->len + len + 1 >= b->cap) {
int nc = MAX(b->len + len + 1, b->cap * 3/2 +1);
b->buf = realloc(b->buf, nc);
b->cap = nc;
}
memcpy(b->buf + b->len, str, len);
b->len += len;
return b; 
}

const char* export buffer_tostring (buffer* b, int* size) {
b->buf[b->len]=0;
if (size) *size = b->len;
return b->buf;
}

static int bnew (lua_State* L) {
unsigned int size = luaL_optint(L,1,32);
buffer* b = luaL_newudata(L, sizeof(buffer), "buffer");
buffer_init(b,size);
return 1;
}

static int bclear (lua_State* L) {
buffer* b = luaL_checkudata(L,1,"buffer");
buffer_clear(b);
lua_settop(L,1);
return 1;
}

static int bfree (lua_State* L) {
buffer* b = luaL_checkudata(L,1,"buffer");
buffer_free(b,0);
return 0;
}

static int append (lua_State* L) {
int len;
buffer* b = luaL_checkudata(L,1,"buffer");
const char* str = luaL_checklstring(L,2,&len);
buffer_append(b, str, len);
lua_settop(L,1);
return 1;
}

static int tostring (lua_State* L) {
buffer* b = luaL_checkudata(L,1,"buffer");
lua_pushlstring(L, b->buf, b->len);
return 1;
}


int export luaopen_buffer (lua_State* L) {
luaL_newclass(L,"buffer");
regt(L, "new", bnew);
regt(L, "append", append);
regt(L, "clear", bclear);
regt(L, "string", tostring);
regt(L, "free", bfree);
regt(L, "__tostring", tostring);
regt(L, "__gc", bfree);
return 1;
}

