#include "global.h"
#include "sha256.h"

#define SWAP(a,b) { typeof(a) c = a; a=b; b=c; }

typedef struct {
unsigned char n, m, p, q;
unsigned char s[256];
unsigned char key[256];
} arcfour;

int export arcfourCrypt (arcfour* r, const unsigned char* in, unsigned char* out, int len) {
if (!r || !in || !out || len<0) return 1;
int i; 
unsigned char c;
for(i=0; i<len; i++) {
r->n++;
r->m += r->s[r->n];
r->p += r->s[r->m] +1;
r->q += r->s[r->p] + 3;
SWAP(r->s[r->n], r->s[r->m]);
SWAP(r->s[r->p], r->s[r->q]);
*(out++) = *(in++) 
^ (r->s[ r->s[r->n] + r->s[r->m] ])
^ (r->s[ r->s[r->p] + r->s[r->q] ]);
}
return 0;
}

int export arcfourInit (arcfour* r, const unsigned char* key, int keylen) {
if (!r || !key || keylen<1 || keylen>256) return 1;
char dummy[1024];
int i=0, j=0;
for (i=0; i<256; i++) r->s[i] = i;
for (i=0; i<256; i++) {
j = (j + r->s[i] + key[i%keylen])&0xFF;
SWAP(r->s[i], r->s[j]);
}
for (i=0; i<256; i+=keylen) {
j = MIN(keylen, 256-i);
if (j<=0) break;
memcpy(r->key +i, key, j);
}
r->p = r->q = r->n = r->m = 0;
arcfourCrypt(r, dummy, dummy, 1024);
return 0;
}

int export encrypt (const char* input, char* output, int length, const char* key, int keylen) {
arcfour r;
return arcfourInit(&r, key, keylen) || arcfourCrypt(&r, input, output, length);
}

int export decrypt (const char* input, char* output, int length, const char* key, int keylen) {
return encrypt(input, output, length, key, keylen);
}

int export sha256Init (CSha256* p) {
if (!p) return 1;
Sha256_Init(p);
return 0;
}

size_t export sha256Update (CSha256* p, const void* buf, size_t len) {
if (!p || !buf) return 0;
Sha256_Update(p, buf, len);
return len;
}

const char* export sha256Finish (CSha256* p) {
if (!p) return NULL;
static char digest[SHA256_DIGEST_SIZE];
Sha256_Final(p, digest);
return digest; 
}

const char* export sha256 (const char* in, size_t len) {
CSha256 c;
if (sha256Init(&c) || sha256Update(&c, in, len)!=len) return NULL;
return sha256Finish(&c);
}
