#include<stdlib.h>
#include<stdio.h>
#include<unistd.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<glob.h>

typedef struct {
int state, pos;
struct glob_t fd;
} fileinfo;

static int findnext (lua_State* L) {
fileinfo* fi = lua_topointer(L,1);
if (fi->status == GLOB_NOMATCH || fi->pos >= fi->fd.gl_pathc) {
globfree(&(fi->fd));
free(fi);
return 0;
}
struct stat st;
const char* path = fi->fd.gl_pathv[fi->pos++];
lua_settop(L,0);
lua_pushstring(L, path);
if (stat(path,&st)) return 1;
lua_pushboolean(L, S_ISDIR(st.st_mode));
lua_pushnumber(L, st.st_mtime);
lua_pushnumber(L, st.st_size);
return 4;
}

static int findfirst (lua_State* L) {
fileinfo* fi = malloc(sizeof(fileinfo));
memset(fi, 0, sizeof(fileinfo));
fi->status = glob(luaL_optstring(L,1,"*"), GLOB_ERR | GLOB_MARK | GLOB_TILDE_CHECK, NULL, &(fi->fd));
if (fi->status && fi->status!=GLOB_NOMATCH) luaL_error(L, "invalid glob pattern or permission denied");
lua_settop(L,0);
lua_getreg(L, "_____findnextfile");
lua_pushlightuserdata(L,fi);
return 2;
}

int export isfile (const char* fn) {
struct stat st;
if (!stat(fn,&st)) return S_ISREG(st.st_mode);
else return 0;
}

int export isdir (const char* fn) {
struct stat st;
if (!stat(fn,&st)) return S_ISDIR(st.st_mode);
else return 0;
}

int export exists (const char* fn) {
return isfile(fn) || isdir(fn);
}

int export fcopy (const char* src, const char* dst, int failIfExists) {
if (failIfExists && isfile(dst)) return 0;
int n;
char buf[4096];
FILE *f1 = fopen(src, "rb"), *f2 = fopen(dst, "wb");
if (!f1 || !f2) {
if (f1) fclose(f1);
if (f2) fclose(f2);
return 0;
}
while ((n = fread(buf, 1, 4096, f1))>0) fwrite(buf, 1, n, f2);
if (ferror(f1) || ferror(f2)) return 0;
fclose(f1);
fclose(f2);
return 1;
}

int export fmove (const char* src, const char* dst) {
return fcopy(src, dst, 1) && remove(src);
}

unsigned int export fsize (const char* fn) {
struct stat st;
if (!stat(fn,&st)) return st.st_size;
else return -1;
}

time_t export ftime (const char* fn, int type) {
struct stat st;
if (!stat(fn,&st)) switch(type){
case 0: return st.st_ctime;
case 1: return st.st_atime;
case 2: return st.st_mtime;
}
else return -1;
}

int export mkdir2 (const char* fn) {
return mkdir(fn, 0777);
}

char* export fullpath (const char* path, char* real, int length) {
return realpath(path, real);
}

