#include "global.h"
#include<lua/lua.h>
#include<lua/lualib.h>
#include<lua/lauxlib.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>

#define LUA_TCDATA 10

static int lua_interstate_passMeta (lua_State* src, int sidx, lua_State* dst, int didx) {
if (luaL_getmetafield(src, sidx, "__name")) {
lua_interstate_copy(dst, src, lua_gettop(src));
lua_pop(src,1);
if (luaL_getmetaname(dst)) lua_setmetatable(dst, didx);
}
return 1;
}

static int lua_interstate_copyUpvalues (lua_State* dst, lua_State* src, int idx) {
int top = lua_gettop(src);
for (int i=1; lua_getupvalue(src, idx, i); i++) {
if (lua_rawequal(src, idx, -1)) lua_pushvalue(dst, -1);
else if (!lua_interstate_copy(dst, src, top+1)) return 0;
lua_setupvalue(dst, -2, i);
lua_pop(src,1);
}
return 1;
}

static int lua_interstate_copyFunction (lua_State* dst, lua_State* src, int idx) {
lua_checkstack(src,2);
lua_getreg(src, "string.dump");
lua_pushvalue(src, idx);
lua_call(src, 1, 1);
int len=0, top = lua_gettop(src);
const char* data = lua_tolstring(src, -1, &len);
if (luaL_loadbuffer(dst, data, len, NULL)) return 0;
if (!lua_interstate_copyUpvalues(dst, src, idx)) return 0;
lua_pop(src,1);
return 1;
}

static int lua_interstate_copyCFunction (lua_State* dst, lua_State* src, int idx) {
void* func = lua_tocfunction(src, idx);
int i, top = lua_gettop(src);
for (i=1; lua_getupvalue(src, idx, i); i++) {
if (lua_rawequal(src, idx, -1)) lua_pushvalue(dst, -1);
else if (!lua_interstate_copy(dst, src, top+1)) return 0;
}
lua_pushcclosure(dst, func, i -1);
return 1;
}

static int lua_interstate_copyFullUserData (lua_State* dst, lua_State* src, int idx) {
int len = lua_objlen(src, idx);
void* udata = lua_newuserdata(dst, len);
memcpy(udata, lua_topointer(src,idx), len);
return lua_interstate_passMeta(src, idx, dst, lua_gettop(dst));
}

static int lua_interstate_copyCdata (lua_State* dst, lua_State* src, int idx) {
lua_pushlightuserdata(dst, *(void**)lua_topointer(src, idx));
return 1;
}

static int lua_interstate_copyTable (lua_State* dst, lua_State* src, int idx) {
int top = lua_gettop(src);
lua_checkstack(src,2);
lua_checkstack(dst, 3);
lua_newtable(dst);
lua_pushnil(src);
while (lua_next(src, idx)) {
if (!lua_interstate_copy(dst, src, top+1) || !lua_interstate_copy(dst, src, top+2)) return 0;
lua_rawset(dst, -3);
lua_pop(src,1);
}
return lua_interstate_passMeta(src, idx, dst, lua_gettop(dst));
}

int lua_interstate_copy (lua_State* dst, lua_State* src, int idx) {
lua_checkstack(dst,1);
switch(lua_type(src,idx)) {
case LUA_TNIL: lua_pushnil(dst); return 1;
case LUA_TBOOLEAN: lua_pushboolean(dst, lua_toboolean(src,idx)); return 1;
case LUA_TNUMBER:  lua_pushnumber(dst, lua_tonumber(src,idx));  return 1;
case LUA_TSTRING: {
int len = 0;
const char* str = lua_tolstring(src, idx, &len);
lua_pushlstring(dst, str, len);
return 1;
}
case LUA_TFUNCTION :
if (lua_iscfunction(src,idx)) return lua_interstate_copyCFunction(dst, src, idx);
else return lua_interstate_copyFunction(dst, src, idx);
case LUA_TTABLE : return lua_interstate_copyTable(dst, src, idx);
case LUA_TLIGHTUSERDATA : lua_pushlightuserdata(dst, lua_topointer(src, idx)); return 1;
case LUA_TUSERDATA : return lua_interstate_copyFullUserData(dst, src, idx);
case LUA_TCDATA : return lua_interstate_copyCdata(dst, src, idx);
default: // unable to copy or unsupported type
lua_pushnil(dst);
return 0; 
}}

