#include "global.h"
#include<lua/lua.h>
#include<lua/lauxlib.h>
#include<lua/lualib.h>
#include<string.h>
#include<stdio.h>
#include<stdlib.h>

static void lua_serializeval (void* b, lua_State* L, int idx) {
switch (lua_type(L,idx)) {
case LUA_TNIL :
buffer_append(b, "E", 1);
break;
case LUA_TBOOLEAN :
buffer_append(b, lua_toboolean(L,idx)?"t":"f", 1);
break;
case LUA_TNUMBER : {
char gbuf[64] = "N";
double d = lua_tonumber(L,idx);
snprintf(gbuf+1, 62, "%.14g", d);
buffer_append(b, gbuf, strlen(gbuf));
}break;
case LUA_TSTRING : {
char gbuf[20] = "S";
size_t len;
const char* str = lua_tolstring(L, idx, &len);
buffer_append(b, itoa(len, 1+gbuf, 10) -1, 0);
if ((*str>='0' && *str<='9') || *str==':') buffer_append(b, ":", 1);
buffer_append(b, str, len);
}break;
case LUA_TTABLE : {
int top = lua_gettop(L);
if (luaL_callmeta(L, idx, "__save")) {
if (lua_isnil(L,top+1)) lua_pop(L,1);
else if (!lua_isnone(L,top+1)) {
lua_replace(L,idx);
lua_serializeval(b, L, idx);
return;
}}
buffer_append(b, "T", 1);
if (luaL_getmetafield(L, idx, "__name")) {
buffer_append(b,"M",1);
lua_serializeval(b, L, top+1);
lua_pop(L,1);
}
lua_checkstack(L,3);
lua_pushnil(L);
while(lua_next(L,idx)) {
lua_serializeval(b, L, top+1);
lua_serializeval(b, L, top+2);
lua_pop(L,1);
}
buffer_append(b, "E", 1);
}break;
case LUA_TFUNCTION : {
lua_getreg(L, "string.dump");
lua_pushvalue(L,idx>0? idx : idx -1);
lua_call(L,1,1);
size_t sl=0;
const char* s = lua_tolstring(L,-1,&sl);
char gbuf[20] = "F";
buffer_append(b, itoa(sl, gbuf+1, 10) -1, 0);
buffer_append(b, s, sl);
lua_pop(L,1);
int i, top = lua_gettop(L);
for (i=1; lua_getupvalue(L,idx,i); i++) {
if (lua_rawequal(L, idx, top+1)) buffer_append(b, "e", 1);
else lua_serializeval(b, L, top+1);
lua_pop(L,1);
}
buffer_append(b, ".", 1);
}break;
case LUA_TLIGHTUSERDATA : {
char ch[9] = "L";
*(void**)(ch+1) = lua_topointer(L,idx);
buffer_append(b, ch, sizeof(void*)+1);
}break;
case LUA_TUSERDATA : {
int top = lua_gettop(L);
if (luaL_callmeta(L, idx, "__save")) {
if (lua_isnil(L,top+1)) lua_pop(L,1);
else if (!lua_isnone(L,top+1)) {
lua_replace(L,idx);
lua_serializeval(b, L, idx);
return;
}}
int len = lua_objlen(L,idx);
char* ptr = lua_topointer(L,idx);
char cb[5] = "U";
*(int*)(cb+1) = len;
buffer_append(b, cb, 5);
if (luaL_getmetafield(L, idx, "__name")) {
buffer_append(b,"M",1);
lua_serializeval(b, L, top+1);
lua_pop(L,1);
}
else buffer_append(b, "-", 1);
buffer_append(b, ptr, len);
}break;
default : {
return luaL_argerror(L, idx, "Couldn't serialize argument");
}break;
}}

static const char* lua_unserializeval (const char* s, lua_State* L) {
switch (*s) {
case 'E' : lua_pushnil(L); s++; break;
case 't' : lua_pushboolean(L,1); s++; break;
case 'f' : lua_pushboolean(L,0); s++; break;
case 'N' :
lua_pushnumber(L, strtod(s+1, &s));
break;
case 'S' : {
size_t len = strtoul(s+1, &s, 10);
lua_pushlstring(L, s+(*s==':'?1:0), len);
s+=len+(*s==':'?1:0);
}break;
case 'T' : {
lua_checkstack(L,3);
lua_newtable(L);
s++;
while (1) {
if (*s=='E') break; 
else if (*s=='M') {
s = lua_unserializeval(++s, L);
if (luaL_getmetaname(L)) lua_setmetatable(L, -2);
continue;
}
s = lua_unserializeval(s, L);
s = lua_unserializeval(s, L);
lua_rawset(L,-3);
}
int top = lua_gettop(L);
if (luaL_callmeta(L, -1, "__load")) {
if (lua_isnil(L,top+1)) lua_pop(L,1);
else if (!lua_isnone(L,top+1)) lua_remove(L,-2);
}
}break;
case 'F' : {
size_t len = strtoul(s+1, &s, 10);
luaL_loadbuffer(L, s, len, NULL);
s += len;
int i=0;
while(1) {
if (*s=='.') {s++; break; }
else if (*s=='e') { s++; lua_pushvalue(L,-1); }
else s = lua_unserializeval(s, L) -1;
lua_setupvalue(L, -2, ++i);
}
}break;
case 'L' : {
lua_pushlightuserdata(L, *(void**)(s+1));
s += sizeof(void*)+1;
}break;
case 'U' : {
int meta=0, len = *(int*)(s+1);
s+=5;
if (*s=='M') {
s = lua_unserializeval(s, L);
meta = luaL_getmetaname(L);
}
char* udata = lua_newuserdata(L, len);
int top = lua_gettop(L);
memcpy(udata, s, len);
s += len;
if (meta) {
lua_pushvalue(L,-2);
lua_setmetatable(L,-2);
lua_remove(L,-2);
}
if (luaL_callmeta(L, -1, "__load")) {
if (lua_isnil(L,top+1)) lua_pop(L,1);
else if (!lua_isnone(L,top+1)) lua_remove(L,-2);
}
}break;
default : {
luaL_error(L, "unknown type specifier '%c'. Unserialize aborted, corrupted data.", *s);
}break;
}
return s;
}

static int serialize (lua_State* L) {
int len;
void* b = buffer_new(64);
for (int i=1, n=lua_gettop(L); i<=n; i++) {
lua_serializeval(b, L, i);
}
lua_settop(L,0);
const char* str = buffer_tostring(b,&len);
lua_pushlstring(L, str, len);
return 1;
}

static int unserialize (lua_State* L) {
int sz, nArgs = 0;
for (int i=1, n=lua_gettop(L); i<=n; i++) {
const char* s = lua_tolstring(L, i, &sz);
while (*s) {
s = lua_unserializeval(s, L);
nArgs++;
}}
return nArgs;
}

int export luaopen_s11n (lua_State* L) {
lua_getglobal(L,"string");
lua_getfield(L, -1, "dump");
lua_setreg(L, "string.dump");
regt(L, "serialize", serialize);
regt(L, "unserialize", unserialize);
return 1;
}
