function tohex (s)
local r = ''
for i = 1, #s do r = r .. string.format('%02x', s:byte(i)) end
return r
end

local ffi = require 'ffi'
local C = ffi.C
ffi.cdef[[
typedef struct { 
unsigned long state[8];
unsigned long long count;
unsigned char buffer[64];
} sha256ctx;
typedef struct {
unsigned char n, m, p, q;
unsigned char state[256];
unsigned char key[256];
} arcfour;
void sha256Init (sha256ctx* ctx) ;
int sha256Update (sha256ctx* ctx, const char* buffer, unsigned int length) ;
const char* sha256Finish (sha256ctx* ctx) ;
int arcfourInit (arcfour* ctx, const char* key, int keylen) ;
int arcfourCrypt (arcfour* ctx, const char* input, char* output, int length) ;
int encrypt (const char* in, char* out, int len, const char* key, int keylen) ;
const char* sha256 (const char* in, int len) ;
]]

function string:encrypt (key, klen) 
local enc = ffi.new('unsigned char[?]', #self)
C.encrypt(self, enc, #self, key, klen or #key)
return ffi.string(enc, #self)
end
string.decrypt = string.encrypt

function string:hash ()
return ffi.string(C.sha256(self, #self), 32)
end

string.sha256 = {
new = function () 
local ctx = ffi.new('sha256ctx') 
C.sha256Init(ctx)
return ctx
end,
update = function (self, str, len) 
return C.sha256Update(self, str, len or #str) 
end,
finish = function (self) return ffi.string(C.sha256Finish(self),32) end,
}

string.encryption  = {
new = function (key, len)
local r = ffi.new('arcfour') 
C.arcfourInit(r, key, len or #key)
return r
end,
crypt = function (self, str, len)
len = len or #str
local out = ffi.new('char[?]', len)
C.arcfourCrypt(self, str, out, len)
return ffi.string(out,len)
end
}
string.encryption.__index = string.encryption
string.sha256.__index = string.sha256
ffi.metatype('sha256ctx', string.sha256)
ffi.metatype('arcfour', string.encryption)

function hex (s)
local r = ''
for i = 1, #s do r = r .. string.format('%02x', s:byte(i)) end
return r
end

local key = 'abcdefgh'
local ctx = string.encryption.new(key, #key)
local e1 = ctx:crypt('Hello')
local e2 = ctx:crypt('World')
local e3 = e1..e2
local re = e3:decrypt(key, #key)
print(type(re), re)


local text = 'HelloWorld'
local text2 = 'Test2'
local ctx = string.sha256.new()
print('re=', ctx:update(text))
ctx:update(text2)
local digest1 = ctx:finish()
local digest2 = string.hash(text..text2)
print('1=', hex(digest1))
print('2=', hex(digest2))
print(digest1==digest2)
